from . import LoadDataMixin, ReperMixin, BaseItemListMixin, SearchableItemMixin
from pyxwb2.utils import manifest


class ShipAbility(LoadDataMixin, ReperMixin):
    pass


class Upgrade(LoadDataMixin, ReperMixin):
    pass


class Upgrades(BaseItemListMixin):
    __singular__ = Upgrade


class DamageCard(LoadDataMixin, ReperMixin):
    pass


class DamageDeck(BaseItemListMixin):
    __singular__ = DamageCard


class Condition(LoadDataMixin, ReperMixin):
    pass


class Conditions(BaseItemListMixin):
    __singular__ = Condition


class ShipStat(LoadDataMixin, ReperMixin):
    pass


class ShipStats(BaseItemListMixin):
    __singular__ = ShipStat


class Action(LoadDataMixin, ReperMixin):
    pass


class Actions(BaseItemListMixin):
    __singular__ = Action


class Maneuver(ReperMixin):
    def __iter__(self):
        return iter(self.__dict__.keys())

    @classmethod
    def load_data(cls, maneuver):
        """
        A single maneuver on a ship's dial

        :param maneuver: 3 character maneuver following xwing-data2 data set which is encoded to match
                        the vassal league nomenclature
        :return: Initialized Maneuver class
        """
        obj = cls()
        speed, direction, difficulty = list(maneuver)
        obj.__setattr__("name", manifest["maneuvers"]["maneuvers"][direction]["name"])
        obj.__setattr__("speed", int(speed))
        obj.__setattr__("difficulty", manifest["maneuvers"]["difficulty"][difficulty]["name"])
        obj.__setattr__("direction", manifest["maneuvers"]["maneuvers"][direction]["direction"])
        obj.__setattr__("vassal", maneuver)

        return obj


class ShipDial(SearchableItemMixin):
    """
    The ship's dial containing all of the possible maneuvers the ship can perform. The returned object
    is a ShipDial with as a list of Maneuver objects.

    This object allows searching a keyword to return a smaller subset of items with that keyword
    e.g. ShipDial['red'] would return a ShipDial object of all maneuvers that have a difficulty of red
         ShipDial['left'] would return a ShipDial object of all maneuvers that have a direction of turn left
         ShipDial['2'] would return a ShipDial object of all maneuvers that speed of 2

    You can also do multi-searches
    e.g. ShipDial['1']['blue'] would return a ShipDial object of all blue difficulty maneuvers with a speed of 1

    However, the object still allows for getting a maneuver at a specific index.
    e.g. ShipDial[1] returns a single maneuver object at the index
    """
    def __init__(self):
        self._items = list()

    def __len__(self):
        return len(self._items)

    def __repr__(self):
        return f"ShipDial({[m for m in self._items][:5]}{' ...' if len(self._items) > 5 else ''})"

    @classmethod
    def load_data(cls, dial):
        obj = cls()
        for maneuver in dial:
            obj._items.append(Maneuver.load_data(maneuver))

        return obj



import jsonschema
import json

from pathlib import Path, PurePath
from jsonpath2 import Path as JPath

from .models.base import Faction, Factions, Ship, Ships
from .models.misc import Actions, Conditions, DamageDeck, ShipStats, Upgrades
from .models.pilot import Pilots, Pilot
from .utils import manifest

_local_path = Path(__file__).parents[0]


def _calculate_cost(squad):
    for pilot in squad.pilots:
        _cost = int()
        try:
            _cost += pilot.cost
        except AttributeError:
            continue

        for upgrade in pilot.upgrades:

            try:
                _cost += upgrade.cost["value"]
            except (KeyError,):
                pass

            try:
                _cost += upgrade.cost["values"][str(getattr(pilot, upgrade.cost["variable"]))]
            except (KeyError, AttributeError):
                pass

            for s in pilot.ship.stats:
                try:
                    stat_value = s.value if s.type == upgrade.cost["variable"] else None
                    _cost += upgrade.cost["values"][stat_value]
                except (KeyError,):
                    continue
        setattr(pilot, "points", _cost)

    setattr(squad, "points", sum([p.points for p in squad.pilots]))


def _load_simple_manifest_data(cls, set_file):
    obj = cls()
    with open(PurePath(_local_path, manifest[set_file]).as_posix(), "r") as f:
        _set = json.load(f)
    for _dc in _set:
        obj.append(obj.__singular__.load_data(_dc))
    return obj


def _load_complex_manifest_data(cls, manifest_set, inner_set=None):
    obj = cls()
    for set_file in manifest[manifest_set]:
        with open(PurePath(_local_path, set_file).as_posix(), "r") as f:
            _set = json.load(f)
        try:
            for _dc in _set[inner_set]:
                obj.append(obj.__singular__.load_data(_dc))
        except (KeyError, TypeError):
            for _dc in _set:
                obj.append(obj.__singular__.load_data(_dc))
    return obj


class XwingSquadron:

    def __init__(self, trust_source=False):
        self.trust_source = trust_source

        self.name = None
        self.description = None
        self.obstacles = None
        self.points = None
        self.vendor = None
        self.version = None
        self.faction = None
        self.pilots = None

    def export_squad(self, as_dict=False):
        """
        Export the squad in XWS format. By default this is exported as a json string.

        :param as_dict: Set to True if you want the returned data as a dict instead of json
        :return: XWS formatted data
        """
        _squad = dict()

        _squad["name"] = getattr(self, "name", "undefined")
        _squad["faction"] = self.faction.xws
        _squad["points"] = self.points
        _squad["version"] = self.version

        _squad["pilots"] = [
            {
                "id": p.xws,
                "points": p.points,
                "upgrades": {
                    s.replace(" ", "").lower(): [
                        u.xws
                        for u in p.upgrades
                        if s in u.sides[0]["slots"]
                    ]
                    for s in set(p.slots)
                }
            }
            for p in self.pilots
        ]

        _squad["vendor"] = {
            "pyxwb2": {
                "builder_url": "https://github.com/minsis/pyxwb2"
            }
        }

        if as_dict:
            return _squad
        return json.dumps(_squad)

    def import_squad(self, xws):
        """
        Import the xws data to validate the squad

        :param xws: Importing data from either an already imported json file, or pass
                    in a string of the json file to process. This will populate the already
                    initialized Xwb object.
        :return: None
        """

        if isinstance(xws, str):
            with open(xws, "r") as xws_file:
                _xws_data = json.load(xws_file)
        elif isinstance(xws, dict):
            _xws_data = xws
        else:
            raise ValueError("Unable to load data set. Be sure to pass in a string of the filename or"
                             "an already loaded dict from the xws json file.")

        # Validate the schema if we dont trust the source. Other validation happens for required items
        # even if this is bypassed.
        if not self.trust_source:
            self._validate_schema(_xws_data)

        self.version = _xws_data.get("version")
        self.name = _xws_data.get("name")
        self.description = _xws_data.get("description")

        self.faction = Faction.load_data(_xws_data.get("faction"))
        self.pilots = Pilots.load_data(_xws_data.get("pilots"), self.faction)

        _calculate_cost(self)

    @staticmethod
    def _validate_schema(xws_data):
        schema_fn = PurePath(_local_path, "data/xws_schema.json").as_posix()
        with open(schema_fn, "r") as f:
            schema = json.load(f)

        jsonschema.validate(instance=xws_data, schema=schema)


class XwingDataPack:

    def __init__(self):
        self.factions = Factions()
        self.pilots = Pilots()
        self.ships = Ships()
        self.actions = Actions()

        self._load()

    def _load(self):
        factions_jpath = JPath.parse_str("$.pilots.*.faction")
        _factions = [m.current_value for m in factions_jpath.match(manifest)]

        for pilot in manifest["pilots"]:
            _faction = Faction.load_data(pilot["faction"])
            self.factions.append(_faction)
            for ship in pilot["ships"]:
                with open(PurePath(_local_path, ship).as_posix(), "r") as f:
                    ship_data = json.load(f)

                _ship = Ship.load_data(ship_data)
                _ship.faction = _faction
                self.ships.append(_ship)

                for pilot_data in ship_data["pilots"]:
                    _pilot = Pilot.load_data(pilot_data)
                    _pilot.__setattr__("faction", _faction)
                    _pilot.__setattr__("ship", _ship)
                    self.pilots.append(_pilot)

        self.actions = _load_complex_manifest_data(Actions, "actions")
        self.conditions = _load_simple_manifest_data(Conditions, "conditions")
        self.damage_deck = _load_complex_manifest_data(DamageDeck, "damagedecks", "cards")
        self.stats = _load_complex_manifest_data(ShipStats, "stats")
        self.upgrades = _load_complex_manifest_data(Upgrades, "upgrades")



from .pyxwb2 import XwingSquadron, XwingDataPack

__version__ = "0.2.0b"
__vendors__ = {
    "xws": "2.0",
    "xwing-data": "1.25.0",
    "dial-maneuvers-conversion": "1.0.0",
    "jsonschema": "2019-09"
}



from pathlib import Path, PurePath
from jsonpath2.path import Path as JPath
import json

from . import BaseItemListMixin
from .misc import ShipDial, Actions, ShipStats
from pyxwb2.utils import manifest

from .exceptions import FactionMissingException


class Faction:
    def __init__(self):
        self.name = None
        self.xws = None
        self.ffg = None

    def __repr__(self):
        return f"Faction(name='{self.name}', xws='{self.xws}'"

    @classmethod
    def load_data(cls, faction_name):
        """
        Create a Faction object by passing in the faction name. This can be either the XWS spec name or,
        the actual name of the faction
        e.g. faction_name = 'rebelalliance'
             faction_name = 'Rebel Alliance'

        :param faction_name: String value of the faction
        :return:
        """
        obj = cls()
        for file in manifest["factions"]:
            filepath = PurePath(Path(__file__).parents[1], file).as_posix()
            with open(filepath, "r") as f:
                factions = json.load(f)

            jpath = JPath.parse_str(f"$..[?(@.xws=\"{faction_name}\"or@.name=\"{faction_name}\")]")
            faction = [m.current_value for m in jpath.match(factions)]

            if not len(faction):
                raise FactionMissingException("Found no valid faction, be sure you have a valid faction "
                                              "or the xws spec is at least version 2.0.0")

            for k, v in faction[0].items():
                obj.__setattr__(k, v)

        return obj


class Factions(BaseItemListMixin):
    __singular__ = Faction

    def __contains__(self, item):
        all_names = [p.xws for p in self._items] + [p.name for p in self._items]
        return item in all_names

    @classmethod
    def load_data(cls, factions):
        obj = cls()

        for faction in factions:
            obj._items.add(Faction.load_data(faction))

        return obj


class Ship:
    def __init__(self):
        self._skip_attr = ["pilots", "actions", "dial", "stats"]
        self.name = None
        self.xws = None
        self.size = None
        self.faction = None

    def __repr__(self):
        return f"Ship(name='{self.name}', xws='{self.xws}', size='{self.size}', faction={self.faction})"

    @classmethod
    def load_data(cls, ship_data):
        obj = cls()
        for k, v in ship_data.items():
            if k in obj._skip_attr:
                continue
            obj.__setattr__(k, v)
        obj.__setattr__("dial", ShipDial.load_data(ship_data["dial"]))
        obj.__setattr__("actions", Actions.load_data(ship_data["actions"]))
        obj.__setattr__("stats", ShipStats.load_data(ship_data["stats"]))

        return obj


class Ships(BaseItemListMixin):
    __singular__ = Ship

    def __contains__(self, item):
        return item in [s.xws for s in self._items] + \
               [s.name for s in self._items]



class LoadDataMixin:
    @classmethod
    def load_data(cls, data):
        obj = cls()
        for k, v in data.items():
            obj.__setattr__(k, v)
        return obj


class BaseItemListMixin:
    def __init__(self, singular=None):
        self._items = list()

        if singular:
            self.__singular__ = singular

    def __len__(self):
        return len(self._items)

    def __getitem__(self, item):
        return self._items[item]

    def __iter__(self):
        return iter(self._items)

    def __repr__(self):
        return f"{self.__class__.__name__}({[f for f in self._items]})"

    def append(self, item):
        if isinstance(item, self.__singular__):
            self._items.append(item)

    @classmethod
    def load_data(cls, data):
        obj = cls()
        for action in data:
            obj._items.append(obj.__singular__.load_data(action))
        return obj


class ReperMixin:
    def __repr__(self):
        def encap_quotes(value):
            if type(value) == str:
                return f"'{value}'"
            return value

        all_attrs = ", ".join([f"{k}={encap_quotes(v)}" for k, v in self.__dict__.items() if not k.startswith('_')])
        return f"{self.__class__.__name__}({all_attrs})"


class SearchableItemMixin:
    _items = list()

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._items[item]
        elif isinstance(item, str):
            _tmp = self.__class__()
            for maneuver in self._items:
                for c in maneuver:
                    if item.lower() in str(getattr(maneuver, c)).lower():
                        _tmp._items.append(maneuver)
            return _tmp



import json
from pathlib import Path, PurePath

manifest_file = PurePath(Path(__file__).parents[0], "data/manifest.json").as_posix()
with open(manifest_file, "r") as f:
    manifest = json.load(f)

bearing_conversion_file = PurePath(Path(__file__).parents[0], "data/dial-maneuvers-conversion.json").as_posix()
with open(bearing_conversion_file) as f:
    manifest["maneuvers"] = json.load(f)



import json
from pathlib import Path, PurePath
from jsonpath2.path import Path as JPath

from pyxwb2.utils import manifest

from . import ReperMixin, BaseItemListMixin
from .base import Ship
from .misc import ShipAbility, Upgrades
from .exceptions import PilotsMissingException


class Pilot(ReperMixin):

    def __init__(self):
        self.slots = None
        self._skip_attr = ["vendor", "alt", "shipAbility"]

    def __contains__(self, item):
        return item in self.__dict__

    @classmethod
    def load_data(cls, pilot_data):
        obj = cls()
        try:
            for k, v in pilot_data.items():
                if k in obj._skip_attr:
                    continue
                obj.__setattr__(k, v)

                if "shipAbility" in pilot_data:
                    obj.__setattr__("ship_ability", ShipAbility.load_data(pilot_data["shipAbility"]))
        except IndexError:
            pass

        return obj


class Pilots(BaseItemListMixin, ReperMixin):
    __singular__ = Pilot

    def __getitem__(self, item):
        return self._items[item]

    def __contains__(self, item):
        all_names = [p.xws for p in self._items] + [p.name for p in self._items]

        return item in all_names

    def export(self):
        _pilots = dict()
        for pilot in self._items:
            _pilots["id"] = pilot.xws



    @classmethod
    def load_data(cls, pilots, faction):
        """
        Create a list of pilot objects loaded from xwing-data.

        This object allows quick searching for pilots that were loaded in. It will also allow you to search
        using a pilot object. This particular search doesn't compare any kind of equality of the Pilot object
        but instead will use the Pilot object and search if the same pilot is contained in Pilots
        e.g. 'bladesquadronveteran' in Pilots
        or   'Blade Squadron Veteran' in Pilots

        :param pilots: Pass in a list of pilots to process. The list should be the direct list of pilots in
                       accordance with the xws spec
        :param faction: Pass in the processed Faction object.
        :return: Processed Pilots object with all incoming pilots attached
        """
        obj = cls()
        jpath_ship = JPath.parse_str(f"$.pilots..[?(@.faction=\"{faction.xws}\")].ships.*")
        manifest_ship_files = [m.current_value for m in jpath_ship.match(manifest)]

        for pilot in pilots:
            pilot_xws = pilot.get("id")
            for file in manifest_ship_files:
                manifest_ship_fn = PurePath(Path(__file__).parents[1], file).as_posix()
                with open(manifest_ship_fn, "r") as f:
                    ship_file_data = json.load(f)
                jpath_pilot = JPath.parse_str(f"$.pilots..[?(@.xws=\"{pilot_xws}\")]")
                pilot_file_data = [m.current_value for m in jpath_pilot.match(ship_file_data)]
                if not pilot_file_data:
                    continue

                _ship = Ship.load_data(ship_file_data)
                _ship.faction = faction
                _pilot = Pilot.load_data(pilot_file_data[0])
                setattr(_pilot, "ship", _ship)
                setattr(_pilot, "faction", faction)

                try:
                    setattr(_pilot, "slots", [s for s in pilot_file_data[0]["slots"]])
                except KeyError:
                    pass

                if "upgrades" in pilot:
                    _upgrades = list()
                    _slots = [s.replace(" ", "").lower() for s in _pilot.slots]
                    for slot, upgrades in pilot["upgrades"].items():
                        if slot not in _slots:
                            continue
                        for fn in manifest["upgrades"]:
                            if slot not in fn:
                                continue
                            manifest_upgrade_fn = PurePath(Path(__file__).parents[1], fn).as_posix()
                            with open(manifest_upgrade_fn, "r") as fn:
                                upgrade_file_data = json.load(fn)
                            for upgrade in upgrades:
                                for ufd in upgrade_file_data:
                                    if upgrade == ufd["xws"]:
                                        _upgrades.append(ufd)
                    setattr(_pilot, "upgrades", Upgrades.load_data(_upgrades))

                try:
                    setattr(_pilot, "points", pilot_file_data[0]["cost"])
                except KeyError:
                    pass

                obj._items.append(_pilot)

        if not len(obj):
            raise PilotsMissingException("Found no valid pilots, be sure you have valid pilots "
                                         "or the xws spec is at least version 2.0.0")
        return obj



from setuptools import setup
from pyxwb2 import __version__ as version

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    requirements = [r for r in f.read().split("\n") if r]


setup(
    name="pyxwb2",
    packages=["pyxwb2", "pyxwb2.models"],
    version=version,
    license="GNU GPLv3",
    description="Python API for FFG X-wing 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dennis Whitney",
    author_email="dennis@irunasroot.com",
    url="https://github.com/minsis/pyxwb2",
    download_url="https://github.com/minsis/pyxwb2/archive/v0.1.0b.tar.gz",
    package_data={
        "pyxwb2": [
            "data/*",
            "data/*/*",
            "data/*/*/*",
            "data/*/*/*/*"
        ]
    },
    keywords=[
      "ffg",
      "x-wing",
      "xwing",
      "x-wing 2.0",
      "xwing 2.0"
    ],
    install_requires=requirements,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.6",
    ]
)
